from bamt.external.pyBN.utils.class_equivalence import *
from bamt.external.pyBN.utils.data import *
from bamt.external.pyBN.utils.graph import *
from bamt.external.pyBN.utils.independence_tests import *



import sys
import warnings
from copy import copy

import numpy as np
import pandas as pd

from bamt.mi_entropy_gauss import mi_gauss as mutual_information, entropy_all as entropy
from bamt.preprocess.graph import edges_to_dict
from bamt.preprocess.numpy_pandas import get_type_numpy


def info_score(edges: list, data: pd.DataFrame, method="LL"):
    score_funcs = {"LL": log_lik_local, "BIC": BIC_local, "AIC": AIC_local}
    score = score_funcs.get(method.upper(), BIC_local)

    parents_dict = edges_to_dict(edges)
    nodes_with_edges = parents_dict.keys()
    scores = [
        score(data[child_parents].copy(), method)
        for var in nodes_with_edges
        for child_parents in ([var] + parents_dict[var],)
    ]
    scores += [
        score(data[[var]].copy(), method)
        for var in set(data.columns).difference(set(nodes_with_edges))
    ]
    return sum(scores)


##### INFORMATION-THEORETIC SCORING FUNCTIONS #####


def log_likelihood(bn, data, method="LL"):
    """
    Determining log-likelihood of the parameters
    of a Bayesian Network. This is a quite simple
    score/calculation, but it is useful as a straight-forward
    structure learning score.

    Semantically, this can be considered as the evaluation
    of the log-likelihood of the data, given the structure
    and parameters of the BN:
            - log( P( D | Theta_G, G ) )
            where Theta_G are the parameters and G is the structure.

    However, for computational reasons it is best to take
    advantage of the decomposability of the log-likelihood score.

    As an example, if you add an edge from A->B, then you simply
    need to calculate LOG(P'(B|A)) - Log(P(B)), and if the value
    is positive then the edge improves the fitness score and should
    therefore be included.

    Even more, you can expand and manipulate terms to calculate the
    difference between the new graph and the original graph as follows:
            Score(G') - Score(G) = M * I(X,Y),
            where M is the number of data points and I(X,Y) is
            the marginal mutual information calculated using
            the empirical distribution over the data.

    In general, the likelihood score decomposes as follows:
            LL(D | Theta_G, G) =
                    M * Sum over Variables ( I ( X , Parents(X) ) ) -
                    M * Sum over Variables ( H( X ) ),
            where 'I' is mutual information and 'H' is the entropy,
            and M is the number of data points

    Moreover, it is clear to see that H(X) is independent of the choice
    of graph structure (G). Thus, we must only determine the difference
    in the mutual information score of the original graph which had a given
    node and its original parents, and the new graph which has a given node
    and new parents.

    NOTE: This assumes the parameters have already
    been learned for the BN's given structure.

    LL = LL - f(N)*|B|, where f(N) = 0

    Arguments
    ---------
    *bn* : a BayesNet object
            Must have both structure and parameters
            instantiated.
    Notes
    -----
    NROW = data.shape[0]
    mi_score = 0
    ent_score = 0
    for rv in bn.nodes():
            cols = tuple([bn.V.index(rv)].extend([bn.V.index(p) for p in bn.parents(rv)]))
            mi_score += mutual_information(data[:,cols])
            ent_score += entropy(data[:,bn.V.index(rv)])

    return NROW * (mi_score - ent_score)
    """

    NROW = data.shape[0]
    mi_scores = [
        mutual_information(
            data[:, (bn.V.index(rv),) + tuple([bn.V.index(p) for p in bn.parents(rv)])],
            method=method,
        )
        for rv in bn.nodes()
    ]
    ent_scores = [entropy(data[:, bn.V.index(rv)], method=method) for rv in bn.nodes()]
    return NROW * (sum(mi_scores) - sum(ent_scores))


def log_lik_local(data, method="LL"):
    NROW = data.shape[0]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if isinstance(data, pd.DataFrame):
            return NROW * (
                mutual_information(data, method=method)
                - entropy(data.iloc[:, 0], method=method)
            )
        elif isinstance(data, pd.Series):
            return 0.0
        elif isinstance(data, np.ndarray):
            return NROW * (
                mutual_information(data, method=method)
                - entropy(data[:, 0], method=method)
            )


def BIC_local(data, method="BIC"):
    NROW = data.shape[0]
    log_score = log_lik_local(data, method=method)
    try:
        penalty = 0.5 * num_params(data) * np.log(NROW)
    except OverflowError as err:
        penalty = sys.float_info.max
    return log_score - penalty


def num_params(data):
    # Convert pandas DataFrame to numpy array
    if isinstance(data, pd.DataFrame):
        data = data.values
    # Convert pandas Series to numpy array
    if isinstance(data, pd.Series):
        data = np.array(copy(data))

    # Calculate number of parameters for numpy array
    if isinstance(data, np.ndarray):
        node_type = get_type_numpy(data)
        columns_for_discrete = [
            param for param, node in node_type.items() if node == "cont"
        ]
        columns_for_code = [
            param for param, node in node_type.items() if node == "disc"
        ]

        prod = 1
        for var in columns_for_code:
            prod *= (
                len(np.unique(data[:, var])) if data.ndim != 1 else len(np.unique(data))
            )
        if columns_for_discrete:
            prod *= len(columns_for_discrete)

        # Handle overflow error
        try:
            return prod
        except OverflowError:
            return sys.float_info.max

    # Raise an error if data type is unexpected
    print("Num_params: Unexpected data type")
    print(data)
    return None


def AIC_local(data, method="AIC"):
    log_score = log_lik_local(data, method=method)
    penalty = num_params(data)
    return log_score - penalty



from datetime import timedelta
from typing import Dict, Optional, List, Tuple, Callable

from golem.core.adapter import DirectAdapter
from golem.core.dag.verification_rules import has_no_cycle, has_no_self_cycled_nodes
from golem.core.log import Log
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.selection import SelectionTypesEnum
from golem.core.optimisers.objective import Objective, ObjectiveEvaluate
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from pandas import DataFrame
from sklearn import preprocessing

import bamt.preprocessors as pp
from bamt.builders.builders_base import VerticesDefiner, EdgesDefiner
from bamt.log import logger_builder
from bamt.nodes.composite_continuous_node import CompositeContinuousNode
from bamt.nodes.composite_discrete_node import CompositeDiscreteNode
from bamt.nodes.discrete_node import DiscreteNode
from bamt.nodes.gaussian_node import GaussianNode
from bamt.utils import EvoUtils as evo
from bamt.utils.composite_utils import CompositeGeneticOperators
from bamt.utils.composite_utils.CompositeModel import CompositeModel, CompositeNode


class CompositeDefiner(VerticesDefiner, EdgesDefiner):
    """
    Object that might take additional methods to decompose structure builder class
    """

    def __init__(
        self,
        descriptor: Dict[str, Dict[str, str]],
        regressor: Optional[object] = None,
    ):
        super().__init__(descriptor, regressor)

        # Notice that vertices are used only by Builders
        self.vertices = []

        # LEVEL 1: Define a general type of node: Discrete or Сontinuous
        for vertex, type in self.descriptor["types"].items():
            if type in ["disc_num", "disc"]:
                node = CompositeDiscreteNode(name=vertex)
            elif type == "cont":
                node = CompositeContinuousNode(name=vertex, regressor=regressor)
            else:
                msg = f"""First stage of automatic vertex detection failed on {vertex} due TypeError ({type}).
                Set vertex manually (by calling set_nodes()) or investigate the error."""
                logger_builder.error(msg)
                continue

            self.vertices.append(node)


class CompositeStructureBuilder(CompositeDefiner):
    """
    This class uses an evolutionary algorithm based on GOLEM to generate a Directed Acyclic Graph (DAG) that represents
    the structure of a Composite Bayesian Network.

    Attributes:
        data (DataFrame): Input data used to build the structure.
        descriptor (dict): Descriptor describing node types and signs.
    """

    def __init__(
        self,
        data: DataFrame,
        descriptor: Dict[str, Dict[str, str]],
        regressor: Optional[object],
    ):
        super(CompositeStructureBuilder, self).__init__(
            descriptor=descriptor, regressor=regressor
        )
        self.data = data
        self.parent_models_dict = {}
        self.descriptor = descriptor
        self.regressor = regressor
        self.params = {
            "init_edges": None,
            "init_nodes": None,
            "remove_init_edges": True,
            "white_list": None,
            "bl_add": None,
        }
        self.default_n_jobs = -1
        self.default_pop_size = 15
        self.default_crossover_prob = 0.9
        self.default_mutation_prob = 0.8
        self.default_max_arity = 100
        self.default_max_depth = 100
        self.default_timeout = 180
        self.default_num_of_generations = 50
        self.default_early_stopping_iterations = 50
        self.objective_metric = CompositeGeneticOperators.composite_metric
        self.default_crossovers = [
            CrossoverTypesEnum.exchange_edges,
            CrossoverTypesEnum.exchange_parents_one,
            CrossoverTypesEnum.exchange_parents_both,
            CompositeGeneticOperators.custom_crossover_all_model,
        ]
        self.default_mutations = [
            CompositeGeneticOperators.custom_mutation_add_structure,
            CompositeGeneticOperators.custom_mutation_delete_structure,
            CompositeGeneticOperators.custom_mutation_reverse_structure,
            CompositeGeneticOperators.custom_mutation_add_model,
        ]
        self.default_selection = [SelectionTypesEnum.tournament]
        self.default_constraints = [
            has_no_self_cycled_nodes,
            has_no_cycle,
            evo.has_no_duplicates,
        ]
        self.verbose = True
        self.logging_level = 50

    def overwrite_vertex(
        self,
        regressor: Optional[Callable],
    ):
        for node_instance in self.vertices:
            node = node_instance
            if (
                len(node_instance.cont_parents + node_instance.disc_parents) < 1
                and type(node_instance).__name__ == "CompositeContinuousNode"
            ):
                node = GaussianNode(name=node_instance.name, regressor=regressor)
            elif (
                len(node_instance.cont_parents + node_instance.disc_parents) < 1
                and type(node_instance).__name__ == "CompositeDiscreteNode"
            ):
                node = DiscreteNode(name=node_instance.name)
            else:
                continue
            id_node = self.skeleton["V"].index(node_instance)
            node.disc_parents = node_instance.disc_parents
            node.cont_parents = node_instance.cont_parents
            node.children = node_instance.children
            self.skeleton["V"][id_node] = node

    def build(
        self,
        data: DataFrame,
        classifier: Optional[object],
        regressor: Optional[object],
        **kwargs,
    ):
        """
        Calls the search method to execute all the evolutionary computations.

        Args:
            data (DataFrame): The data from which to build the structure.
            classifier (Optional[object]): A classification model for discrete nodes.
            regressor (Optional[object]): A regression model for continuous nodes.
        """
        best_graph_edge_list, parent_models = self.search(data, **kwargs)

        # Convert the best graph to the format used by the Bayesian Network
        self.skeleton["V"] = self.vertices
        self.skeleton["E"] = best_graph_edge_list
        self.parent_models_dict = parent_models

        self.get_family()
        self.overwrite_vertex(regressor=regressor)

    def search(self, data: DataFrame, **kwargs) -> [List[Tuple[str, str]], Dict]:
        """
        Executes all the evolutionary computations and returns the best graph's edge list.

        Args:
            data (DataFrame): The data from which to build the structure.

        Returns:
            best_graph_edge_list (List[Tuple[str, str]]): The edge list of the best graph found by the search.
        """
        # Get the list of node names
        vertices = list(data.columns)

        encoder = preprocessing.LabelEncoder()
        p = pp.Preprocessor([("encoder", encoder)])
        preprocessed_data, _ = p.apply(data)

        # Create the initial population
        initial = kwargs.get(
            "custom_initial_structure",
            [
                CompositeModel(
                    nodes=[
                        CompositeNode(
                            nodes_from=None,
                            content={
                                "name": vertex,
                                "type": p.nodes_types[vertex],
                                "parent_model": None,
                            },
                        )
                        for vertex in vertices
                    ]
                )
            ],
        )

        # Define the requirements for the evolutionary algorithm
        requirements = GraphRequirements(
            max_arity=kwargs.get("max_arity", self.default_max_arity),
            max_depth=kwargs.get("max_depth", self.default_max_depth),
            num_of_generations=kwargs.get(
                "num_of_generations", self.default_num_of_generations
            ),
            timeout=timedelta(minutes=kwargs.get("timeout", self.default_timeout)),
            early_stopping_iterations=kwargs.get(
                "early_stopping_iterations", self.default_early_stopping_iterations
            ),
            n_jobs=kwargs.get("n_jobs", self.default_n_jobs),
        )

        # Set the parameters for the evolutionary algorithm
        optimizer_parameters = GPAlgorithmParameters(
            pop_size=kwargs.get("pop_size", self.default_pop_size),
            crossover_prob=kwargs.get("crossover_prob", self.default_crossover_prob),
            mutation_prob=kwargs.get("mutation_prob", self.default_mutation_prob),
            genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
            mutation_types=kwargs.get("custom_mutations", self.default_mutations),
            crossover_types=kwargs.get("custom_crossovers", self.default_crossovers),
            selection_types=kwargs.get("selection_type", self.default_selection),
        )

        # Set the adapter for the conversion between the graph and the data
        # structures used by the optimizer
        adapter = DirectAdapter(
            base_graph_class=CompositeModel, base_node_class=CompositeNode
        )

        # Set the constraints for the graph

        constraints = kwargs.get("custom_constraints", [])

        constraints.extend(self.default_constraints)

        if kwargs.get("blacklist", None) is not None:
            constraints.append(evo.has_no_blacklist_edges)
        if kwargs.get("whitelist", None) is not None:
            constraints.append(evo.has_only_whitelist_edges)

        graph_generation_params = GraphGenerationParams(
            adapter=adapter, rules_for_constraint=constraints
        )

        # Define the objective function to optimize
        objective = Objective(
            {"custom": kwargs.get("custom_metric", self.objective_metric)}
        )

        # Initialize the optimizer
        optimizer = EvoGraphOptimizer(
            objective=objective,
            initial_graphs=initial,
            requirements=requirements,
            graph_generation_params=graph_generation_params,
            graph_optimizer_params=optimizer_parameters,
        )

        # Define the function to evaluate the objective function
        objective_eval = ObjectiveEvaluate(objective, data=preprocessed_data)

        if not kwargs.get("verbose", self.verbose):
            Log().reset_logging_level(logging_level=50)

        # Run the optimization
        optimized_graph = optimizer.optimise(objective_eval)[0]

        parent_models = self._get_parent_models(optimized_graph)

        # Get the best graph
        best_graph_edge_list = optimized_graph.operator.get_edges()
        best_graph_edge_list = self._convert_to_strings(best_graph_edge_list)

        return best_graph_edge_list, parent_models

    @staticmethod
    def _convert_to_strings(nested_list):
        return [[str(item) for item in inner_list] for inner_list in nested_list]

    @staticmethod
    def _get_parent_models(graph):
        parent_models = {}
        for node in graph.nodes:
            parent_models[node.content["name"]] = node.content["parent_model"]
        return parent_models



import random
from typing import Optional, List, Union

import numpy as np
from pandas import DataFrame
from sklearn import linear_model

from .base import BaseNode
from .schema import LogitParams


class LogitNode(BaseNode):
    """
    Main class for logit node
    """

    def __init__(self, name, classifier: Optional[object] = None):
        super(LogitNode, self).__init__(name)
        if classifier is None:
            classifier = linear_model.LogisticRegression(
                multi_class="multinomial", solver="newton-cg", max_iter=100
            )
        self.classifier = classifier
        self.type = "Logit" + f" ({type(self.classifier).__name__})"

    def fit_parameters(self, data: DataFrame, **kwargs) -> LogitParams:
        parents = self.cont_parents + self.disc_parents
        self.classifier.fit(X=data[parents].values, y=data[self.name].values, **kwargs)

        return {
            "classes": list(self.classifier.classes_),
            "classifier_obj": self.classifier,
            "classifier": type(self.classifier).__name__,
            "serialization": None,
        }

    def get_dist(self, node_info, pvals):
        if len(node_info["classes"]) > 1:
            model = node_info["classifier_obj"]
            if type(self).__name__ == "CompositeDiscreteNode":
                pvals = [int(item) if isinstance(item, str) else item for item in pvals]

            return model.predict_proba(np.array(pvals).reshape(1, -1))[0]
        else:
            return np.array([1.0])

    def choose(self, node_info: LogitParams, pvals: List[Union[float]]) -> str:
        """
        Return value from Logit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        rindex = 0

        distribution = self.get_dist(node_info, pvals)

        if len(node_info["classes"]) > 1:
            rand = random.random()
            lbound = 0
            ubound = 0
            for interval in range(len(node_info["classes"])):
                ubound += distribution[interval]
                if lbound <= rand < ubound:
                    rindex = interval
                    break
                else:
                    lbound = ubound

            return str(node_info["classes"][rindex])
        else:
            return str(node_info["classes"][0])

    @staticmethod
    def predict(node_info: LogitParams, pvals: List[Union[float]]) -> str:
        """
        Return prediction from Logit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        if len(node_info["classes"]) > 1:
            model = node_info["classifier_obj"]
            pred = model.predict(np.array(pvals).reshape(1, -1))[0]

            return str(pred)

        else:
            return str(node_info["classes"][0])



from typing import Optional, Union, List

from golem.core.dag.graph_delegate import GraphDelegate
from golem.core.dag.linked_graph_node import LinkedGraphNode


class CompositeNode(LinkedGraphNode):
    def __str__(self):
        return self.content["name"]


class CompositeModel(GraphDelegate):
    def __init__(
        self, nodes: Optional[Union[LinkedGraphNode, List[LinkedGraphNode]]] = None
    ):
        super().__init__(nodes)
        self.unique_pipeline_id = 1



from typing import Dict, List, Any, Union, TypedDict, Optional

from numpy import ndarray


class DiscreteParams(TypedDict):
    cprob: Union[List[Union[list, Any]], Dict[str, list]]
    vals: List[str]


class MixtureGaussianParams(TypedDict):
    mean: List[float]
    coef: List[float]
    covars: List[float]


class GaussianParams(TypedDict):
    regressor: str
    regressor_obj: Optional[object]
    variance: Union[ndarray, float]
    mean: Union[ndarray, float]
    serialization: Optional[str]


class CondGaussParams(TypedDict):
    regressor: str
    regressor_obj: Optional[object]
    variance: Union[ndarray, float]
    mean: Union[ndarray, float]
    serialization: Optional[str]


class CondMixtureGaussParams(TypedDict):
    mean: Optional[List[float]]
    coef: List[float]
    covars: Optional[List[float]]


class LogitParams(TypedDict):
    classes: List[int]
    classifier: str
    classifier_obj: Optional[object]
    serialization: Optional[str]


class HybcprobParams(TypedDict):
    hybcprob: Dict[str, CondGaussParams]



import math

import numpy as np
import pandas as pd
from scipy import stats
from scipy.stats.distributions import chi2
from sklearn.metrics import mutual_info_score
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import OrdinalEncoder


def lrts_comp(data):
    n = 0
    biggets_p = -1 * np.infty
    comp_biggest = 0
    max_comp = 10
    if len(data) < max_comp:
        max_comp = len(data)
    for i in range(1, max_comp + 1, 1):
        gm1 = GaussianMixture(n_components=i, random_state=0)
        gm2 = GaussianMixture(n_components=i + 1, random_state=0)
        gm1.fit(data)
        ll1 = np.mean(gm1.score_samples(data))
        gm2.fit(data)
        ll2 = np.mean(gm2.score_samples(data))
        LR = 2 * (ll2 - ll1)
        p = chi2.sf(LR, 1)
        if p > biggets_p:
            biggets_p = p
            comp_biggest = i
        n = comp_biggest
    return n


def mix_norm_cdf(x, weights, means, covars):
    mcdf = 0.0
    for i in range(len(weights)):
        mcdf += weights[i] * stats.norm.cdf(x, loc=means[i][0], scale=covars[i][0][0])
    return mcdf


def theoretical_quantile(data, n_comp):
    model = GaussianMixture(n_components=n_comp, random_state=0)
    model.fit(data)
    q = []
    x = []
    # step =  ((np.max(model.sample(100000)[0])) - (np.min(model.sample(100000)[0])))/1000
    step = (np.max(data) - np.min(data)) / 1000
    d = np.arange(np.min(data), np.max(data), step)
    for i in d:
        x.append(i)
        q.append(mix_norm_cdf(i, model.weights_, model.means_, model.covariances_))
    return x, q


def quantile_mix(p, vals, q):
    ind = q.index(min(q, key=lambda x: abs(x - p)))
    return vals[ind]


def probability_mix(val, vals, q):
    ind = vals.index(min(vals, key=lambda x: abs(x - val)))
    return q[ind]


def sum_dist(data, vals, q):
    percs = np.linspace(1, 100, 10)
    x = np.quantile(data, percs / 100)
    y = []
    for p in percs:
        y.append(quantile_mix(p / 100, vals, q))
    dist = 0
    for xi, yi in zip(x, y):
        dist = dist + (abs(-1 * xi + yi)) / math.sqrt(2)
    return dist


def component(data, columns, method):
    n = 1
    max_comp = 10
    x = []
    if data.shape[0] < max_comp:
        max_comp = data.shape[0]
    if len(columns) == 1:
        x = np.transpose([data[columns[0]].values])
    else:
        x = data[columns].values
    if method == "aic":
        lowest_aic = np.infty
        comp_lowest = 0
        for i in range(1, max_comp + 1, 1):
            gm1 = GaussianMixture(n_components=i, random_state=0)
            gm1.fit(x)
            aic1 = gm1.aic(x)
            if aic1 < lowest_aic:
                lowest_aic = aic1
                comp_lowest = i
            n = comp_lowest

    if method == "bic":
        lowest_bic = np.infty
        comp_lowest = 0
        for i in range(1, max_comp + 1, 1):
            gm1 = GaussianMixture(n_components=i, random_state=0)
            gm1.fit(x)
            bic1 = gm1.bic(x)
            if bic1 < lowest_bic:
                lowest_bic = bic1
                comp_lowest = i
            n = comp_lowest

    if method == "LRTS":
        n = lrts_comp(x)
    if method == "quantile":
        biggest_p = -1 * np.infty
        comp_biggest = 0
        for i in range(1, max_comp, 1):
            vals, q = theoretical_quantile(x, i)
            dist = sum_dist(x, vals, q)
            p = probability_mix(dist, vals, q)
            if p > biggest_p:
                biggest_p = p
                comp_biggest = i
        n = comp_biggest
    return n


def get_n_nearest(data, columns, corr=False, number_close=5):
    """Returns N nearest neighbors for every column of dataframe, added into list

    Args:
        data (DataFrame): Proximity matrix
        columns (list): df.columns.tolist()
        corr (bool, optional): _description_. Defaults to False.
        number_close (int, optional): Number of nearest neighbors. Defaults to 5.

    Returns:
        groups
    """
    groups = []
    for c in columns:
        if corr:
            close_ind = data[c].sort_values(ascending=False).index.tolist()
        else:
            close_ind = data[c].sort_values().index.tolist()
        groups.append(close_ind[0 : number_close + 1])

    return groups


def get_proximity_matrix(df, proximity_metric) -> pd.DataFrame:
    """Returns matrix of proximity matrix of the dataframe, dataframe must be coded first if it contains
                                                                                                    categorical data

    Args:
        df (DataFrame): data
        proximity_metric (str): 'MI' or 'corr'

    Returns:
        df_distance: mutual information matrix
    """

    encoder = OrdinalEncoder()
    df_coded = df
    columnsToEncode = list(df_coded.select_dtypes(include=["category", "object"]))

    df_coded[columnsToEncode] = encoder.fit_transform(df_coded[columnsToEncode])

    df_distance = pd.DataFrame(
        data=np.zeros((len(df.columns), len(df.columns))), columns=df.columns
    )
    df_distance.index = df.columns

    if proximity_metric == "MI":
        for c1 in df.columns:
            for c2 in df.columns:
                dist = mutual_info_score(df_coded[c1].values, df_coded[c2].values)
                df_distance.loc[c1, c2] = dist

    elif proximity_metric == "corr":
        df_distance = df_coded.corr(method="pearson")

    return df_distance


def get_brave_matrix(df_columns, proximity_matrix, n_nearest=5) -> pd.DataFrame:
    """Returns matrix Brave coeffitients of the DataFrame, requires proximity measure to be calculated

    Args:
        df_columns (DataFrame): data.columns
        proximity_matrix (DataFrame): may be generated by get_mutual_info_score_matrix() function or
                                                                                                correlation from scipy
        n_nearest (int, optional): _description_. Defaults to 5.

    Returns:
        brave_matrix: DataFrame of Brave coefficients
    """

    brave_matrix = pd.DataFrame(
        data=np.zeros((len(df_columns), len(df_columns))), columns=df_columns
    )
    brave_matrix.index = df_columns

    groups = get_n_nearest(
        proximity_matrix, df_columns.tolist(), corr=True, number_close=n_nearest
    )

    counter_zeroer = 0.0

    for c1 in df_columns:
        for c2 in df_columns:
            a = counter_zeroer
            b = counter_zeroer
            c = counter_zeroer
            d = counter_zeroer
            if c1 != c2:
                for g in groups:
                    if (c1 in g) & (c2 in g):
                        a += 1
                    if (c1 in g) & (c2 not in g):
                        b += 1
                    if (c1 not in g) & (c2 in g):
                        c += 1
                    if (c1 not in g) & (c2 not in g):
                        d += 1

                if (a + c) * (b + d) != 0 and (a + b) * (c + d) != 0:
                    br = (a * len(groups) + (a + c) * (a + b)) / (
                        (math.sqrt((a + c) * (b + d))) * (math.sqrt((a + b) * (c + d)))
                    )
                else:
                    br = (a * len(groups) + (a + c) * (a + b)) / 0.0000000001
                brave_matrix.loc[c1, c2] = br

    return brave_matrix


def _child_dict(net: list):
    res_dict = dict()
    for e0, e1 in net:
        if e1 in res_dict:
            res_dict[e1].append(e0)
        else:
            res_dict[e1] = [e0]
    return res_dict


def precision_recall(pred_net: list, true_net: list, decimal=4):
    pred_dict = _child_dict(pred_net)
    true_dict = _child_dict(true_net)
    corr_undirected = 0
    corr_dir = 0
    for e0, e1 in pred_net:
        flag = True
        if e1 in true_dict:
            if e0 in true_dict[e1]:
                corr_undirected += 1
                corr_dir += 1
                flag = False
        if (e0 in true_dict) and flag:
            if e1 in true_dict[e0]:
                corr_undirected += 1
    pred_len = len(pred_net)
    true_len = len(true_net)
    shd = pred_len + true_len - corr_undirected - corr_dir
    return {
        "AP": round(corr_undirected / pred_len, decimal),
        "AR": round(corr_undirected / true_len, decimal),
        #        'F1_undir': round(2 * (corr_undirected / pred_len) * (corr_undirected / true_len) / (corr_undirected / pred_len + corr_undirected / true_len), decimal),
        "AHP": round(corr_dir / pred_len, decimal),
        "AHR": round(corr_dir / true_len, decimal),
        # 'F1_directed': round(2*(corr_dir/pred_len)*(corr_dir/true_len)/(corr_dir/pred_len+corr_dir/true_len), decimal),
        "SHD": shd,
    }



import math
import random
from typing import Optional, List

import numpy as np
from pandas import DataFrame
from sklearn import linear_model
from sklearn.metrics import mean_squared_error as mse

from .base import BaseNode
from .schema import GaussianParams


class GaussianNode(BaseNode):
    """
    Main class for Gaussian Node
    """

    def __init__(self, name, regressor: Optional[object] = None):
        super(GaussianNode, self).__init__(name)
        if regressor is None:
            regressor = linear_model.LinearRegression()
        self.regressor = regressor
        self.type = "Gaussian" + f" ({type(self.regressor).__name__})"

    def fit_parameters(self, data: DataFrame, **kwargs) -> GaussianParams:
        parents = self.cont_parents
        if type(self).__name__ == "CompositeContinuousNode":
            parents = parents + self.disc_parents
        if parents:
            self.regressor.fit(data[parents].values, data[self.name].values, **kwargs)
            predicted_value = self.regressor.predict(data[parents].values)
            variance = mse(data[self.name].values, predicted_value, squared=False)
            return {
                "mean": np.nan,
                "regressor_obj": self.regressor,
                "regressor": type(self.regressor).__name__,
                "variance": variance,
                "serialization": None,
            }
        else:
            mean_base = np.mean(data[self.name].values)
            variance = np.var(data[self.name].values)
            return {
                "mean": mean_base,
                "regressor_obj": None,
                "regressor": None,
                "variance": variance,
                "serialization": None,
            }

    def get_dist(self, node_info, pvals):
        var = node_info["variance"]
        if pvals:
            for el in pvals:
                if str(el) == "nan":
                    return np.nan
            model = node_info["regressor_obj"]

            if type(self).__name__ == "CompositeContinuousNode":
                pvals = [int(item) if isinstance(item, str) else item for item in pvals]

            cond_mean = model.predict(np.array(pvals).reshape(1, -1))[0]
            return cond_mean, var
        else:
            return node_info["mean"], math.sqrt(var)

    def choose(self, node_info: GaussianParams, pvals: List[float]) -> float:
        """
        Return value from Logit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        cond_mean, var = self.get_dist(node_info, pvals)
        return random.gauss(cond_mean, var)

    @staticmethod
    def predict(node_info: GaussianParams, pvals: List[float]) -> float:
        """
        Return prediction from Logit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        if pvals:
            for el in pvals:
                if str(el) == "nan":
                    return np.nan
            model = node_info["regressor_obj"]
            pred = model.predict(np.array(pvals).reshape(1, -1))[0]
            return pred
        else:
            return node_info["mean"]



from typing import Optional, Union

from sklearn import linear_model

from .gaussian_node import GaussianNode
from .schema import GaussianParams, HybcprobParams

NodeInfo = Union[GaussianParams, HybcprobParams]


class CompositeContinuousNode(GaussianNode):
    def __init__(self, name, regressor: Optional[object] = None):
        super(CompositeContinuousNode, self).__init__(name)
        if regressor is None:
            regressor = linear_model.LinearRegression()
        self.regressor = regressor
        self.type = "CompositeContinuous" + f" ({type(self.regressor).__name__})"



from typing import Union, List, Optional

import numpy as np
from gmr import GMM
from pandas import DataFrame

from bamt.utils.MathUtils import component
from .base import BaseNode
from .schema import MixtureGaussianParams


class MixtureGaussianNode(BaseNode):
    """
    Main class for Mixture Gaussian Node
    """

    def __init__(self, name):
        super(MixtureGaussianNode, self).__init__(name)
        self.type = "MixtureGaussian"

    def fit_parameters(self, data: DataFrame) -> MixtureGaussianParams:
        """
        Train params for Mixture Gaussian Node
        """
        parents = self.disc_parents + self.cont_parents
        if not parents:
            n_comp = int(
                (
                    component(data, [self.name], "aic")
                    + component(data, [self.name], "bic")
                )
                / 2
            )  # component(data, [node], 'LRTS')#
            # n_comp = 3
            gmm = GMM(n_components=n_comp).from_samples(
                np.transpose([data[self.name].values]),
                n_iter=500,
                init_params="kmeans++",
            )
            means = gmm.means.tolist()
            cov = gmm.covariances.tolist()
            # weigts = np.transpose(gmm.to_responsibilities(np.transpose([data[node].values])))
            w = gmm.priors.tolist()  # []
            # for row in weigts:
            #     w.append(np.mean(row))
            return {"mean": means, "coef": w, "covars": cov}
        if parents:
            if not self.disc_parents and self.cont_parents:
                nodes = [self.name] + self.cont_parents
                new_data = data[nodes]
                new_data.reset_index(inplace=True, drop=True)
                n_comp = int(
                    (
                        component(new_data, nodes, "aic")
                        + component(new_data, nodes, "bic")
                    )
                    / 2
                )  # component(new_data, nodes, 'LRTS')#
                # n_comp = 3
                gmm = GMM(n_components=n_comp).from_samples(
                    new_data[nodes].values, n_iter=500, init_params="kmeans++"
                )
                means = gmm.means.tolist()
                cov = gmm.covariances.tolist()
                # weigts = np.transpose(gmm.to_responsibilities(new_data[nodes].values))
                w = gmm.priors.tolist()  # []
                # for row in weigts:
                #     w.append(np.mean(row))
                return {"mean": means, "coef": w, "covars": cov}

    @staticmethod
    def get_dist(node_info, pvals):
        mean = node_info["mean"]
        covariance = node_info["covars"]
        w = node_info["coef"]
        n_comp = len(node_info["coef"])
        if n_comp != 0:
            if pvals:
                indexes = [i for i in range(1, len(pvals) + 1)]
                if not np.isnan(np.array(pvals)).all():
                    gmm = GMM(
                        n_components=n_comp,
                        priors=w,
                        means=mean,
                        covariances=covariance,
                    )
                    cond_gmm = gmm.condition(indexes, [pvals])
                    return cond_gmm.means, cond_gmm.covariances, cond_gmm.priors
                else:
                    return np.nan, np.nan, np.nan
            else:
                gmm = GMM(
                    n_components=n_comp, priors=w, means=mean, covariances=covariance
                )
                return gmm.means, gmm.covariances, gmm.priors
        else:
            return np.nan, np.nan, np.nan

    def choose(
        self, node_info: MixtureGaussianParams, pvals: List[Union[str, float]]
    ) -> Optional[float]:
        """
        Func to get value from current node
        node_info: nodes info from distributions
        pvals: parent values
        Return value from MixtureGaussian node
        """
        mean, covariance, w = self.get_dist(node_info, pvals)

        n_comp = len(w)

        gmm = GMM(
            n_components=n_comp,
            priors=w,
            means=mean,
            covariances=covariance,
        )
        return gmm.sample(1)[0][0]

    @staticmethod
    def predict(
        node_info: MixtureGaussianParams, pvals: List[Union[str, float]]
    ) -> Optional[float]:
        """
        Func to get prediction from current node
        node_info: nodes info from distributions
        pvals: parent values
        Return value from MixtureGaussian node
        """
        mean = node_info["mean"]
        covariance = node_info["covars"]
        w = node_info["coef"]
        n_comp = len(node_info["coef"])
        if n_comp != 0:
            if pvals:
                indexes = [i for i in range(1, len(pvals) + 1)]
                if not np.isnan(np.array(pvals)).all():
                    gmm = GMM(
                        n_components=n_comp,
                        priors=w,
                        means=mean,
                        covariances=covariance,
                    )
                    sample = gmm.predict(indexes, [pvals])[0][0]
                else:
                    sample = np.nan
            else:
                # gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                sample = 0
                for ind, wi in enumerate(w):
                    sample += wi * mean[ind][0]
        else:
            sample = np.nan
        return sample



"""
******************************
Conditional Independence Tests
for Constraint-based Learning
******************************

Implemented Constraint-based Tests
----------------------------------
- mutual information
- Pearson's X^2

I may consider putting this code into its own class structure. The
main benefit I could see from doing this would be the ability to
cache joint/marginal/conditional probabilities for expedited tests.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np

from bamt.external.pyBN.utils.data import unique_bins


def mutual_information(data, conditional=False):
    # bins = np.amax(data, axis=0)+1 # read levels for each variable
    bins = unique_bins(data)
    if len(bins) == 1:
        hist, _ = np.histogramdd(data, bins=(bins))  # frequency counts
        Px = (hist / hist.sum()) + (1e-7)
        MI = -1 * np.sum(Px * np.log(Px))
        return round(MI, 4)

    if len(bins) == 2:
        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        Pxy = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Px = np.sum(Pxy, axis=1)  # P(X,Z)
        Py = np.sum(Pxy, axis=0)  # P(Y,Z)

        PxPy = np.outer(Px, Py)
        Pxy += 1e-7
        PxPy += 1e-7
        MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
        return round(MI, 4)
    elif len(bins) > 2 and conditional:
        # CHECK FOR > 3 COLUMNS -> concatenate Z into one column
        if len(bins) > 3:
            data = data.astype("str")
            ncols = len(bins)
            for i in range(len(data)):
                data[i, 2] = "".join(data[i, 2:ncols])
            data = data.astype(np.int64)[:, 0:3]

        bins = np.amax(data, axis=0)
        hist, _ = np.histogramdd(data, bins=bins)  # frequency counts

        Pxyz = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Pz = np.sum(Pxyz, axis=(0, 1))  # P(Z)
        Pxz = np.sum(Pxyz, axis=1)  # P(X,Z)
        Pyz = np.sum(Pxyz, axis=0)  # P(Y,Z)

        Pxy_z = Pxyz / (Pz + 1e-7)  # P(X,Y | Z) = P(X,Y,Z) / P(Z)
        Px_z = Pxz / (Pz + 1e-7)  # P(X | Z) = P(X,Z) / P(Z)
        Py_z = Pyz / (Pz + 1e-7)  # P(Y | Z) = P(Y,Z) / P(Z)

        Px_y_z = np.empty((Pxy_z.shape))  # P(X|Z)P(Y|Z)
        for i in range(bins[0]):
            for j in range(bins[1]):
                for k in range(bins[2]):
                    Px_y_z[i][j][k] = Px_z[i][k] * Py_z[j][k]
        Pxyz += 1e-7
        Pxy_z += 1e-7
        Px_y_z += 1e-7
        MI = np.sum(Pxyz * np.log(Pxy_z / (Px_y_z)))

        return round(MI, 4)
    elif len(bins) > 2 and conditional == False:
        data = data.astype("str")
        ncols = len(bins)
        for i in range(len(data)):
            data[i, 1] = "".join(data[i, 1:ncols])
        data = data.astype(np.int64)[:, 0:2]

        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        Pxy = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Px = np.sum(Pxy, axis=1)  # P(X,Z)
        Py = np.sum(Pxy, axis=0)  # P(Y,Z)

        PxPy = np.outer(Px, Py)
        Pxy += 1e-7
        PxPy += 1e-7
        MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
        return round(MI, 4)


def entropy(data):
    """
    In the context of structure learning, and more specifically
    in constraint-based algorithms which rely on the mutual information
    test for conditional independence, it has been proven that the variable
    X in a set which MAXIMIZES mutual information is also the variable which
    MINIMIZES entropy. This fact can be used to reduce the computational
    requirements of tests based on the following relationship:

        Entropy is related to marginal mutual information as follows:
            MI(X;Y) = H(X) - H(X|Y)

        Entropy is related to conditional mutual information as follows:
            MI(X;Y|Z) = H(X|Z) - H(X|Y,Z)

        For one varibale, H(X) is equal to the following:
            -1 * sum of p(x) * log(p(x))

        For two variables H(X|Y) is equal to the following:
            sum over x,y of p(x,y)*log(p(y)/p(x,y))

        For three variables, H(X|Y,Z) is equal to the following:
            -1 * sum of p(x,y,z) * log(p(x|y,z)),
                where p(x|y,z) = p(x,y,z)/p(y)*p(z)
    Arguments
    ----------
    *data* : a nested numpy array
        The data from which to learn - must have at least three
        variables. All conditioned variables (i.e. Z) are compressed
        into one variable.

    Returns
    -------
    *H* : entropy value

    """
    try:
        cols = data.shape[1]
    except IndexError:
        cols = 1

    bins = np.amax(data, axis=0)
    if isinstance(bins, np.ndarray):
        for i in range(len(bins)):
            if bins[i] == 0:
                bins[i] = 1
    else:
        bins = 1
    # bins = unique_bins(data)

    if cols == 1:
        hist, _ = np.histogramdd(data, bins=(bins))  # frequency counts
        Px = hist / hist.sum() + (1e-7)
        H = -1 * np.sum(Px * np.log(Px))

    elif cols == 2:  # two variables -> assume X then Y
        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        Pxy = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Py = np.sum(Pxy, axis=0)  # P(Y)
        Py += 1e-7
        Pxy += 1e-7
        H = np.sum(Pxy * np.log(Py / Pxy))

    else:
        # CHECK FOR > 3 COLUMNS -> concatenate Z into one column
        if cols > 3:
            data = data.astype("str")
            ncols = len(bins)
            for i in range(len(data)):
                data[i, 2] = "".join(data[i, 2:ncols])
            data = data.astype(np.int64)[:, 0:3]

        bins = np.amax(data, axis=0)
        hist, _ = np.histogramdd(data, bins=bins)  # frequency counts

        Pxyz = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Pyz = np.sum(Pxyz, axis=0)

        Pxyz += 1e-7  # for log -inf
        Pyz += 1e-7
        H = -1 * np.sum(Pxyz * np.log(Pxyz)) + np.sum(Pyz * np.log(Pyz))

    return round(H, 4)



"""
Collection of Graph Algorithms.

Any networkx dependencies should exist here only, but
there are currently a few exceptions which will
eventually be corrected.

"""

import networkx as nx


def would_cause_cycle(e, u, v, reverse=False):
    """
    Test if adding the edge u -> v to the BayesNet
    object would create a DIRECTED (i.e. illegal) cycle.
    """
    G = nx.DiGraph(e)
    if reverse:
        G.remove_edge(v, u)
    G.add_edge(u, v)
    try:
        nx.find_cycle(G, source=u)
        return True
    except BaseException:
        return False


def topsort(edge_dict, root=None):
    """
    List of nodes in topological sort order from edge dict
    where key = rv and value = list of rv's children
    """
    queue = []
    if root is not None:
        queue = [root]
    else:
        for rv in edge_dict.keys():
            prior = True
            for p in edge_dict.keys():
                if rv in edge_dict[p]:
                    prior = False
            if prior:
                queue.append(rv)

    visited = []
    while queue:
        vertex = queue.pop(0)
        if vertex not in visited:
            visited.append(vertex)
            for nbr in edge_dict[vertex]:
                queue.append(nbr)
            # queue.extend(edge_dict[vertex]) # add all vertex's children
    return visited



"""
********
UnitTest
OrientEdges
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import os
import unittest
from os.path import dirname

import numpy as np
from external.pyBN.structure_learn.orient_edges import orient_edges_gs, orient_edges_pc


class OrientEdgesTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_orient_edges_pc(self):
        e = {0: [1, 2], 1: [0], 2: [0]}
        b = {0: [], 1: {2: (0,)}, 2: {1: (0,)}}
        self.assertDictEqual(orient_edges_pc(e, b), {0: [1, 2], 1: [], 2: []})

    def test_orient_edges_gs(self):
        e = {0: [1, 2], 1: [0], 2: [0]}
        b = {0: [1, 2], 1: [0], 2: [0]}
        dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))), "data")
        path = os.path.join(dpath, "lizards.csv")
        data = np.loadtxt(path, dtype="int32", skiprows=1, delimiter=",")
        self.assertDictEqual(
            orient_edges_gs(e, b, data, 0.05), {0: [1, 2], 1: [], 2: []}
        )



"""
********
UnitTest
BayesNet
********

Method	Checks that
assertEqual(a, b)	a == b
assertNotEqual(a, b)	a != b
assertTrue(x)	bool(x) is True
assertFalse(x)	bool(x) is False
assertIs(a, b)	a is b
assertIsNot(a, b)	a is not b
assertIsNone(x)	x is None
assertIsNotNone(x)	x is not None
assertIn(a, b)	a in b
assertNotIn(a, b)	a not in b
assertIsInstance(a, b)	isinstance(a, b)
assertNotIsInstance(a, b)	not isinstance(a, b)

assertAlmostEqual(a, b)	round(a-b, 7) == 0
assertNotAlmostEqual(a, b)	round(a-b, 7) != 0
assertGreater(a, b)	a > b
assertGreaterEqual(a, b)	a >= b
assertLess(a, b)	a < b
assertLessEqual(a, b)	a <= b

assertListEqual(a, b)	lists
assertTupleEqual(a, b)	tuples
assertSetEqual(a, b)	sets or frozensets
assertDictEqual(a, b)	dicts

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import os
import unittest
from os.path import dirname

from external.pyBN.classes.bayesnet import BayesNet
from external.pyBN.readwrite.read import read_bn


class BayesNetTestCase(unittest.TestCase):
    def setUp(self):
        self.bn = BayesNet()
        self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))), "data")
        self.bn_bif = read_bn(os.path.join(self.dpath, "cancer.bif"))
        self.bn_bn = read_bn(os.path.join(self.dpath, "cmu.bn"))

    def tearDown(self):
        pass

    def test_isinstance(self):
        self.assertIsInstance(self.bn, BayesNet)

    def test_V_bif(self):
        self.assertListEqual(
            self.bn_bif.V, ["Smoker", "Pollution", "Cancer", "Xray", "Dyspnoea"]
        )

    def test_E_bif(self):
        self.assertDictEqual(
            self.bn_bif.E,
            {
                "Cancer": ["Xray", "Dyspnoea"],
                "Dyspnoea": [],
                "Pollution": ["Cancer"],
                "Smoker": ["Cancer"],
                "Xray": [],
            },
        )

    def test_F_bif(self):
        self.assertDictEqual(
            self.bn_bif.F,
            {
                "Cancer": {
                    "cpt": [0.03, 0.97, 0.05, 0.95, 0.001, 0.999, 0.02, 0.98],
                    "parents": ["Pollution", "Smoker"],
                    "values": ["True", "False"],
                },
                "Dyspnoea": {
                    "cpt": [0.65, 0.35, 0.3, 0.7],
                    "parents": ["Cancer"],
                    "values": ["True", "False"],
                },
                "Pollution": {
                    "cpt": [0.9, 0.1],
                    "parents": [],
                    "values": ["low", "high"],
                },
                "Smoker": {
                    "cpt": [0.3, 0.7],
                    "parents": [],
                    "values": ["True", "False"],
                },
                "Xray": {
                    "cpt": [0.9, 0.1, 0.2, 0.8],
                    "parents": ["Cancer"],
                    "values": ["positive", "negative"],
                },
            },
        )

    def test_V_bn(self):
        self.assertListEqual(
            self.bn_bn.V, ["Burglary", "Earthquake", "Alarm", "JohnCalls", "MaryCalls"]
        )

    def test_E_bn(self):
        self.assertDictEqual(
            self.bn_bn.E,
            {
                "Alarm": ["JohnCalls", "MaryCalls"],
                "Burglary": ["Alarm"],
                "Earthquake": ["Alarm"],
                "JohnCalls": [],
                "MaryCalls": [],
            },
        )

    def test_F_bn(self):
        self.assertDictEqual(
            self.bn_bn.F,
            {
                "Alarm": {
                    "cpt": [0.999, 0.001, 0.71, 0.29, 0.06, 0.94, 0.05, 0.95],
                    "parents": ["Earthquake", "Burglary"],
                    "values": ["No", "Yes"],
                },
                "Burglary": {
                    "cpt": [0.999, 0.001],
                    "parents": [],
                    "values": ["No", "Yes"],
                },
                "Earthquake": {
                    "cpt": [0.998, 0.002],
                    "parents": [],
                    "values": ["No", "Yes"],
                },
                "JohnCalls": {
                    "cpt": [0.95, 0.05, 0.1, 0.9],
                    "parents": ["Alarm"],
                    "values": ["No", "Yes"],
                },
                "MaryCalls": {
                    "cpt": [0.99, 0.01, 0.3, 0.7],
                    "parents": ["Alarm"],
                    "values": ["No", "Yes"],
                },
            },
        )

    def test_nodes(self):
        n = list(self.bn_bn.nodes())
        self.assertListEqual(
            n, ["Burglary", "Earthquake", "Alarm", "JohnCalls", "MaryCalls"]
        )

    def test_cpt(self):
        cpt = list(self.bn_bn.cpt("Alarm"))
        self.assertListEqual(cpt, [0.999, 0.001, 0.71, 0.29, 0.06, 0.94, 0.05, 0.95])

    def test_card(self):
        self.assertEqual(self.bn_bn.card("Alarm"), 2)

    def test_scope(self):
        self.assertListEqual(
            self.bn_bn.scope("Alarm"), ["Alarm", "Earthquake", "Burglary"]
        )

    def test_parents(self):
        self.assertListEqual(self.bn_bn.parents("Alarm"), ["Earthquake", "Burglary"])

    def test_values(self):
        self.assertListEqual(self.bn_bn.values("Alarm"), ["No", "Yes"])

    def test_values_idx(self):
        self.assertEqual(self.bn_bn.values("Alarm")[1], "Yes")


if __name__ == "__main__":
    unittest.main(exit=False)
